;;;;; topfun.lsp (pvm-version) ;;;;;

(define compile '())
(define compile-function '())
(define set-handler-stack! '())  ;;; By N.Watanabe
(define compile-system '())      ;;;;; By A243  Nov, 1992 ;;;;;
(define get-name '())

(define *fun-data* '())        ;;;;; By A243  Nov, 1992 ;;;;;
(define *global-symbol* '())
(define pvm-parent '())
(define pvm-recv '())
(define pvm-upk '())
(define pvm-exit '())

(define *_init_scheme* #f)
(define *_init_common* #t)

(define (sctop)
  (let ((input nil))
    (if *_init_common*
        (let ()
          (load '../lsp/common)
          (load '../lsp/help)
          (load '../lsp/err_system)
          (set! *_init_common* #f)
          (set! *_init_scheme* #t))
        (if *_init_scheme*
            (begin
              (if (or (file-exists? "scheme.lsp") (file-exists? "scheme.fas"))
		  (load 'scheme)
		  (let* ((scm-dir (getenv "SCM_DIR"))
			 (path "scheme")
			 (exist-file '()))
		    (if (not (zero? (string-length scm-dir)))
			(begin
			  (set! path (string-append scm-dir "/" path))
			  (set! exist-file
			    (or (and (file-exists? (string-append path ".fas"))
				     ".fas")
				(and (file-exists? (string-append path ".lsp"))
				     ".lsp")))
			  (if exist-file
			      (load (string-append path exist-file)))))))
              (set! *_init_scheme* #f))))
    (set-handler-stack! '())      ;;; By N.Watanabe 
    (set! *debugger-level* 0)     ;;;   for error handler and
    (set! *dbg-cont-stack* '())   ;;;       debugger.
    (set! *condition-stack* '())  ;;;
    (do () (#f)
      (set! *trace-level* 0)
      (display #\newline)
      (display "SC> ")
      (set! input (read))
      (if (not (eof-object? input))
          (write (eval input))
          (begin (display #\newline)
            (eval '(bye)))))))

;; SCTOP for -e option --- read-eval loop
(define (eosctop . strport)
  (let ((strport (if (null? strport)
                     (console-input-port)
                     (car strport)))
        (input nil))
    (set! *loading-message* #f)
    (if (or (file-exists? "scheme.lsp") (file-exists? "scheme.fas"))
        (load 'scheme)
        (let* ((scm-dir (getenv "SCM_DIR"))
               (path "scheme")
               (exist-file '()))
          (if (not (zero? (string-length scm-dir)))
              (begin
                (set! path (string-append scm-dir "/" path))
                (set! exist-file
                      (or (and (file-exists? (string-append path ".fas"))
                               ".fas")
                          (and (file-exists? (string-append path ".lsp"))
                               ".lsp")))
                (if exist-file
                  (load (string-append path exist-file)))))))
    (set-handler-stack! '())      ;;; By N.Watanabe 
    (set! *debugger-level* 0)     ;;;   for error handler and
    (set! *dbg-cont-stack* '())   ;;;       debugger.
    (set! *condition-stack* '())  ;;;
    (do () (#f)
      (set! *trace-level* 0)
      (set! input (read strport))
      (if (not (eof-object? input))
          (eval input)
          (eval '(exit-tus))))))

(define (eval exp)                   ;;;;; By A243  Dec,1992 ;;;;;
  (set! *fun-data* '())              
  (cond ((symbol? exp)
	     (get-symbol-value exp))
        ((pair? exp)
         (cond ((= *re-compile-level* 3)
                (let* ((source (compile exp))
                       (result (bctran-exec-form source source))
                       (symbol-name 
                         (case (car exp)
                           ((define macro set!)
                            (set-pvm-global-symbol result exp))
                           (else #f))))
                  (if (and symbol-name 
                        (check-global-symbol *global-symbol* symbol-name))
                      (set! *global-symbol* (cons symbol-name *global-symbol*)))
                  (eval-answer exp result)))
               ((>= *re-compile-level* 1)
                (let* ((lambda-source (get-lambda-source exp)) 
                      (fun-name (get-name exp))
                      (check-com (if fun-name (macro-function fun-name) #f))
                      (result (bctran-exec-form (compile exp) lambda-source)))
                  (compile-system exp fun-name check-com) 
                  (eval-answer exp result)))
               (else
                 (let ((result (bctran-exec-form (compile exp) '())))
                   (eval-answer exp result))))) 
        (else exp)))

(define (set-pvm-global-symbol result exp)
  (let ((x (if (symbol? result) (eval result) result))) ;;; T.Komiya
    (if (and (function? x) 
             (null? (get-fun-name x)))
        (put-fun-name x (symbol->string (cadr exp))))
    (if (symbol? (cadr exp))
        (cadr exp)
        (caadr exp))))

(define (eval-answer exp result)
  (case (car exp)
    ((define macro)
     (if (pair? (cadr exp))
       (caadr exp)
       (cadr exp)))
    (else result)))

(define (get-lambda-source exp)   
  (case (car exp)
          ((define macro)
             (if (symbol? (cadr exp)) (caddr exp)
               (cons 'lambda (cons (cdadr exp) (cddr exp)))))
          (else '())))

(define (check-global-symbol global-symbol symbol-name)
  (let ((answer #t))
    (for-each (lambda (x) (if (equal? x symbol-name)
                            (set! answer #f))) global-symbol)
    answer))


(define (define-a-function first args body)
  (bctran-define-form (compile-function first args body))
  (car args))


;(define (exec-exp exp)
;  (let* ((bc (compile-function 'define '(top-level) (list exp)))
;	 (first (car bc))
;	 (body (caddr bc))
;	 (func (apply bctran-bcbody
;	              first
;		      (cadr (cadr body))	;;; name
;		      (caddr body)		;;; arg-info
;		      (cadr (cadddr body))	;;; byte code list
;		      () )))
;   (user-function-apply func ())))


(define (top-level-error . message)
  (print2 "top-level-error: ")
  (apply print2 message)
  (error-break))


(define aaa (protect-symbol 'xreturn))
(define aaa (protect-symbol 'lreturn))
(define aaa (protect-symbol 'xcall))
(define aaa (protect-symbol 'make-nil-closure))

;;;;;;;;;;;;;;;;;;;;;;;;; By A243 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (remote-eval)
  (pvm-recv (pvm-parent) #f)
  (set! *global-symbol* (pvm-upk))
  (let* ((recv-data (pvm-upk))
         (environment (car recv-data))
         (global-symbol-value (cdr recv-data)))
    (for-each (lambda (x)
      (if (car x)
        (bctran-exec-form (child-define (car x) (cdr x)) '())
        (if (function? (cddr x))
          (put-closure-env (cddr x) (cadr x)))))
      global-symbol-value)
    (let* ((source (pvm-upk))
           (pre-execute (bctran-exec-form source source))
           (execute (if (symbol? pre-execute)
                      (eval pre-execute)
                      pre-execute)))
      (put-closure-env execute environment)
      (put-fun-name execute (pvm-upk))
      (put-closure-stad execute (pvm-upk))
      (put-closure-arginfo execute (pvm-upk))
      (apply execute (pvm-upk)))
    (pvm-exit)))

(define (child-define x y)
    `(define ,x
       (load-time-eval '((,(car y) ,(cdr y)) (xreturn)))))

(define (remote-eval2)
  (pvm-recv (pvm-parent) #f)
  (set! *global-symbol* (pvm-upk))
  (let* ((recv-data (pvm-upk))
         (environment (car recv-data))
         (global-symbol-value (cdr recv-data)))
    (for-each (lambda (x)
      (if (car x)
        (bctran-exec-form (child-define (car x) (cdr x)) '())
        (if (function? (cddr x))
          (put-closure-env (cddr x) (cadr x)))))
      global-symbol-value)
    (let* ((source (pvm-upk))
           (pre-execute (bctran-exec-form source source))
           (execute (if (symbol? pre-execute)
                      (eval pre-execute)
                      pre-execute)))
      (put-closure-env execute environment)
      (put-fun-name execute (pvm-upk))
      (put-closure-stad execute (pvm-upk))
      (put-closure-arginfo execute (pvm-upk))
      (pvm-recv (pvm-parent) #f)
      (apply execute (pvm-upk)))
    (pvm-exit)))
